@@ -54,28 +54,23 @@
 #include <simgear/bucket/newbucket.hxx>
 #include <simgear/misc/sg_path.hxx>
 #include <simgear/misc/strutils.hxx>
+#include <simgear/misc/sgstream.hxx>
 #include <simgear/threads/SGQueue.hxx>
+#include <simgear/threads/SGThread.hxx>
+#include <simgear/threads/SGGuard.hxx>
+
 #include <simgear/misc/sg_dir.hxx>
 #include <simgear/debug/BufferedLogCallback.hxx>
 #include <simgear/props/props_io.hxx>
 #include <simgear/io/HTTPClient.hxx>
-#include <simgear/io/SVNRepository.hxx>
 #include <simgear/io/HTTPRepository.hxx>
 #include <simgear/io/DNSClient.hxx>
 #include <simgear/structure/exception.hxx>
 #include <simgear/math/sg_random.h>
 
-static const bool svn_built_in_available = true;
-
 using namespace simgear;
 using namespace std;
 
-const char* rsync_cmd =
-        "rsync --verbose --archive --delete --perms --owner --group";
-
-const char* svn_options =
-        "checkout -q";
-
 namespace UpdateInterval
 {
     // interval in seconds to allow an update to repeat after a successful update (=daily)
@@ -89,6 +84,7 @@ typedef map<string,time_t> TileAgeCache;
 ///////////////////////////////////////////////////////////////////////////////
 // helper functions ///////////////////////////////////////////////////////////
 ///////////////////////////////////////////////////////////////////////////////
+
 string stripPath(string path)
 {
     // svn doesn't like trailing white-spaces or path separators - strip them!
@@ -162,16 +158,17 @@ class SyncSlot
 public:
     SyncSlot() :
         isNewDirectory(false),
-        busy(false)
+        busy(false),
+        pendingKBytes(0)
     {}
 
     SyncItem currentItem;
     bool isNewDirectory;
     std::queue<SyncItem> queue;
-    std::auto_ptr<AbstractRepository> repository;
+    std::auto_ptr<HTTPRepository> repository;
     SGTimeStamp stamp;
     bool busy; ///< is the slot working or idle
-
+    unsigned int pendingKBytes;
     unsigned int nextWarnTimeout;
 };
 
@@ -199,101 +196,126 @@ static unsigned int syncSlotForType(SyncItem::Type ty)
     }
 }
 
-///////////////////////////////////////////////////////////////////////////////
-// Base server query
-///////////////////////////////////////////////////////////////////////////////
-
-class ServerSelectQuery : public HTTP::Request
+struct TerrasyncThreadState
 {
-public:
-    ServerSelectQuery() :
-        HTTP::Request("http://scenery.flightgear.org/svn-server", "GET")
-    {
-    }
+    TerrasyncThreadState() :
+        _busy(false),
+        _stalled(false),
+        _fail_count(0),
+        _updated_tile_count(0),
+        _success_count(0),
+        _consecutive_errors(0),
+        _allowed_errors(6),
+        _cache_hits(0),
+        _transfer_rate(0),
+        _total_kb_downloaded(0),
+        _totalKbPending(0)
+    {}
 
-    std::string svnUrl() const
-    {
-        std::string s = simgear::strutils::strip(m_url);
-        if (s.at(s.length() - 1) == '/') {
-            s = s.substr(0, s.length() - 1);
-        }
+    bool _busy;
+    bool _stalled;
+    int  _fail_count;
+    int  _updated_tile_count;
+    int  _success_count;
+    int  _consecutive_errors;
+    int  _allowed_errors;
+    int  _cache_hits;
+    int _transfer_rate;
+    // kbytes, not bytes, because bytes might overflow 2^31
+    int _total_kb_downloaded;
+    unsigned int _totalKbPending;
 
-        return s;
-    }
-protected:
-    virtual void gotBodyData(const char* s, int n)
-    {
-        m_url.append(std::string(s, n));
-    }
-
-    virtual void onFail()
-    {
-        SG_LOG(SG_TERRASYNC, SG_ALERT, "Failed to query TerraSync SVN server");
-        HTTP::Request::onFail();
-    }
-
-private:
-    std::string m_url;
 };
 
-
 ///////////////////////////////////////////////////////////////////////////////
-// SGTerraSync::SvnThread /////////////////////////////////////////////////////
+// SGTerraSync::WorkerThread //////////////////////////////////////////////////
 ///////////////////////////////////////////////////////////////////////////////
-class SGTerraSync::SvnThread : public SGThread
+class SGTerraSync::WorkerThread : public SGThread
 {
 public:
-   SvnThread();
-   virtual ~SvnThread( ) { stop(); }
+   WorkerThread();
+   virtual ~WorkerThread( ) { stop(); }
 
    void stop();
    bool start();
 
-    bool isIdle() {return !_busy; }
+    bool isIdle()
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        return !_state._busy;
+    }
+
+    bool isRunning()
+    {
+         SGGuard<SGMutex> g(_stateLock);
+        return _running;
+    }
+
+    bool isStalled()
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        return _state._stalled;
+    }
+
    void request(const SyncItem& dir) {waitingTiles.push_front(dir);}
-   bool isDirty() { bool r = _is_dirty;_is_dirty = false;return r;}
-   bool hasNewTiles() { return !_freshTiles.empty();}
+
+    bool hasNewTiles()
+    {
+        return !_freshTiles.empty();
+    }
+
    SyncItem getNewTile() { return _freshTiles.pop_front();}
 
-   void   setSvnServer(string server)       { _svn_server   = stripPath(server);}
-   void   setSvnDataServer(string server)   { _svn_data_server   = stripPath(server);}
    void   setHTTPServer(const std::string& server)
    {
       _httpServer = stripPath(server);
    }
 
-   void   setExtSvnUtility(string svn_util) { _svn_command  = simgear::strutils::strip(svn_util);}
-   void   setRsyncServer(string server)     { _rsync_server = simgear::strutils::strip(server);}
+   void setDNSDN( const std::string & dn )
+   {
+     _dnsdn = simgear::strutils::strip(dn);
+   }
+
+   void setSceneryVersion( const std::string & sceneryVersion )
+   {
+     _sceneryVersion = simgear::strutils::strip(sceneryVersion);
+   }
+
    void   setLocalDir(string dir)           { _local_dir    = stripPath(dir);}
    string getLocalDir()                     { return _local_dir;}
-   void   setUseSvn(bool use_svn)           { _use_svn = use_svn;}
-   void   setAllowedErrorCount(int errors)  {_allowed_errors = errors;}
+
+    void setInstalledDir(const SGPath& p)  { _installRoot = p; }
+
+   void   setAllowedErrorCount(int errors)
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        _state._allowed_errors = errors;
+    }
 
    void   setCachePath(const SGPath& p)     {_persistentCachePath = p;}
-   void   setCacheHits(unsigned int hits)   {_cache_hits = hits;}
-   void   setUseBuiltin(bool built_in) { _use_built_in = built_in;}
-
-   volatile bool _active;
-   volatile bool _running;
-   volatile bool _busy;
-   volatile bool _stalled;
-   volatile int  _fail_count;
-   volatile int  _updated_tile_count;
-   volatile int  _success_count;
-   volatile int  _consecutive_errors;
-   volatile int  _allowed_errors;
-   volatile int  _cache_hits;
-   volatile int _transfer_rate;
-   // kbytes, not bytes, because bytes might overflow 2^31
-   volatile int _total_kb_downloaded;
+   void   setCacheHits(unsigned int hits)
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        _state._cache_hits = hits;
+    }
 
+    TerrasyncThreadState threadsafeCopyState()
+    {
+        TerrasyncThreadState st;
+        {
+            SGGuard<SGMutex> g(_stateLock);
+            st = _state;
+        }
+        return st;
+    }
 private:
-   virtual void run();
+    void incrementCacheHits()
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        _state._cache_hits++;
+    }
 
-    // external model run and helpers
-    void runExternal();
-    void syncPathExternal(const SyncItem& next);
-    bool runExternalSyncCommand(const char* dir);
+   virtual void run();
 
     // internal mode run and helpers
     void runInternal();
@@ -308,76 +330,64 @@ private:
     void fail(SyncItem failedItem);
     void notFound(SyncItem notFoundItem);
 
-    bool _use_built_in;
     HTTP::Client _http;
     SyncSlot _syncSlots[NUM_SYNC_SLOTS];
 
-   volatile bool _is_dirty;
-   volatile bool _stop;
+    bool _stop, _running;
    SGBlockingDeque <SyncItem> waitingTiles;
 
    TileAgeCache _completedTiles;
    TileAgeCache _notFoundItems;
 
    SGBlockingDeque <SyncItem> _freshTiles;
-   bool _use_svn;
-   string _svn_server;
-   string _svn_data_server;
-   string _svn_command;
-   string _rsync_server;
    string _local_dir;
    SGPath _persistentCachePath;
    string _httpServer;
+    SGPath _installRoot;
+   string _sceneryVersion;
+   string _dnsdn;
+
+    TerrasyncThreadState _state;
+    SGMutex _stateLock;
 };
 
-SGTerraSync::SvnThread::SvnThread() :
-    _active(false),
-    _running(false),
-    _busy(false),
-    _stalled(false),
-    _fail_count(0),
-    _updated_tile_count(0),
-    _success_count(0),
-    _consecutive_errors(0),
-    _allowed_errors(6),
-    _cache_hits(0),
-    _transfer_rate(0),
-    _total_kb_downloaded(0),
-    _use_built_in(true),
-    _is_dirty(false),
+SGTerraSync::WorkerThread::WorkerThread() :
     _stop(false),
-    _use_svn(true)
+    _running(false)
 {
     _http.setUserAgent("terrascenery-" SG_STRINGIZE(SIMGEAR_VERSION));
 }
 
-void SGTerraSync::SvnThread::stop()
+void SGTerraSync::WorkerThread::stop()
 {
     // drop any pending requests
     waitingTiles.clear();
 
-    if (!_running)
+    if (!isRunning())
         return;
 
     // set stop flag and wake up the thread with an empty request
-    _stop = true;
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        _stop = true;
+    }
+
     SyncItem w(string(), SyncItem::Stop);
     request(w);
     join();
-    _running = false;
 }
 
-bool SGTerraSync::SvnThread::start()
+bool SGTerraSync::WorkerThread::start()
 {
-    if (_running)
+    if (isRunning())
         return false;
 
     if (_local_dir=="")
     {
         SG_LOG(SG_TERRASYNC,SG_ALERT,
                "Cannot start scenery download. Local cache directory is undefined.");
-        _fail_count++;
-        _stalled = true;
+        _state._fail_count++;
+        _state._stalled = true;
         return false;
     }
 
@@ -387,8 +397,8 @@ bool SGTerraSync::SvnThread::start()
         SG_LOG(SG_TERRASYNC,SG_ALERT,
                "Cannot start scenery download. Directory '" << _local_dir <<
                "' does not exist. Set correct directory path or create directory folder.");
-        _fail_count++;
-        _stalled = true;
+        _state._fail_count++;
+        _state._stalled = true;
         return false;
     }
 
@@ -398,45 +408,15 @@ bool SGTerraSync::SvnThread::start()
         SG_LOG(SG_TERRASYNC,SG_ALERT,
                "Cannot start scenery download. Directory '" << _local_dir <<
                "' contains the base package. Use a separate directory.");
-        _fail_count++;
-        _stalled = true;
+        _state._fail_count++;
+        _state._stalled = true;
         return false;
     }
 
-    _use_svn |= _use_built_in;
-
-
-    if ((!_use_svn)&&(_rsync_server==""))
-    {
-        SG_LOG(SG_TERRASYNC,SG_ALERT,
-               "Cannot start scenery download. Rsync scenery server is undefined.");
-        _fail_count++;
-        _stalled = true;
-        return false;
-    }
-
-    _fail_count = 0;
-    _updated_tile_count = 0;
-    _success_count = 0;
-    _consecutive_errors = 0;
     _stop = false;
-    _stalled = false;
-    _running = true;
+    _state = TerrasyncThreadState(); // clean state
 
-    string status;
-
-    if (_use_svn && _use_built_in)
-        status = "Using built-in SVN support. ";
-    else if (_use_svn)
-    {
-        status = "Using external SVN utility '";
-        status += _svn_command;
-        status += "'. ";
-    }
-    else
-    {
-        status = "Using RSYNC. ";
-    }
+    string status = "Using built-in HTTP support.";
 
     // not really an alert - but we want to (always) see this message, so user is
     // aware we're downloading scenery (and using bandwidth).
@@ -449,70 +429,19 @@ bool SGTerraSync::SvnThread::start()
     return true;
 }
 
-bool SGTerraSync::SvnThread::runExternalSyncCommand(const char* dir)
+void SGTerraSync::WorkerThread::run()
 {
-    ostringstream buf;
-    SGPath localPath( _local_dir );
-    localPath.append( dir );
-
-    if (_use_svn)
     {
-        buf << "\"" << _svn_command << "\" "
-            << svn_options << " "
-            << "\"" << _svn_server << "/" << dir << "\" "
-            << "\"" << localPath.str_native() << "\"";
-    } else {
-        buf << rsync_cmd << " "
-            << "\"" << _rsync_server << "/" << dir << "/\" "
-            << "\"" << localPath.str_native() << "/\"";
+         SGGuard<SGMutex> g(_stateLock);
+        _running = true;
     }
 
-    string command;
-#ifdef SG_WINDOWS
-        // windows command line parsing is just lovely...
-        // to allow white spaces, the system call needs this:
-        // ""C:\Program Files\something.exe" somearg "some other arg""
-        // Note: whitespace strings quoted by a pair of "" _and_ the
-        //       entire string needs to be wrapped by "" too.
-        // The svn url needs forward slashes (/) as a path separator while
-        // the local path needs windows-native backslash as a path separator.
-    command = "\"" + buf.str() + "\"";
-#else
-    command = buf.str();
-#endif
-    SG_LOG(SG_TERRASYNC,SG_DEBUG, "sync command '" << command << "'");
-
-#ifdef SG_WINDOWS
-    // tbd: does Windows support "popen"?
-    int rc = system( command.c_str() );
-#else
-    FILE* pipe = popen( command.c_str(), "r");
-    int rc=-1;
-    // wait for external process to finish
-    if (pipe)
-        rc = pclose(pipe);
-#endif
-
-    if (rc)
-    {
-        SG_LOG(SG_TERRASYNC,SG_ALERT,
-               "Failed to synchronize directory '" << dir << "', " <<
-               "error code= " << rc);
-        return false;
-    }
-    return true;
-}
-
-void SGTerraSync::SvnThread::run()
-{
-    _active = true;
     initCompletedTilesPersistentCache();
 
     if (_httpServer == "automatic" ) {
 
-        //TODO: make DN and service settable from properties
-        DNS::NAPTRRequest * naptrRequest = new DNS::NAPTRRequest("terrasync.flightgear.org");
-        naptrRequest->qservice = "ws20";
+        DNS::NAPTRRequest * naptrRequest = new DNS::NAPTRRequest(_dnsdn);
+        naptrRequest->qservice = _sceneryVersion;
 
         naptrRequest->qflags = "U";
         DNS::Request_ptr r(naptrRequest);
@@ -566,106 +495,17 @@ void SGTerraSync::SvnThread::run()
           SG_LOG(SG_TERRASYNC,SG_INFO, "picking entry # " << idx << ", server is " << _httpServer );
         }
     }
-    if( _httpServer.empty() ) { // don't resolve SVN server is HTTP server is set
-        if ( _svn_server.empty()) {
-            SG_LOG(SG_TERRASYNC,SG_INFO, "Querying closest TerraSync server");
-            ServerSelectQuery* ssq = new ServerSelectQuery;
-            HTTP::Request_ptr req = ssq;
-            _http.makeRequest(req);
-            while (!req->isComplete()) {
-                _http.update(20);
-            }
 
-            if (req->readyState() == HTTP::Request::DONE) {
-                _svn_server = ssq->svnUrl();
-                SG_LOG(SG_TERRASYNC,SG_INFO, "Closest TerraSync server:" << _svn_server);
-            } else {
-                SG_LOG(SG_TERRASYNC,SG_WARN, "Failed to query closest TerraSync server");
-            }
-        } else {
-            SG_LOG(SG_TERRASYNC,SG_INFO, "Explicit: TerraSync server:" << _svn_server);
-        }
+    runInternal();
 
-        if (_svn_server.empty()) {
-#if WE_HAVE_A_DEFAULT_SERVER_BUT_WE_DONT_THIS_URL_IS_NO_LONGER_VALID
-            // default value
-            _svn_server = "http://foxtrot.mgras.net:8080/terrascenery/trunk/data/Scenery";
-#endif
-        }
-    }
-
-    if (_use_built_in) {
-        runInternal();
-    } else {
-        runExternal();
-    }
-
-    _active = false;
-    _running = false;
-    _is_dirty = true;
-}
-
-void SGTerraSync::SvnThread::runExternal()
-{
-    while (!_stop)
     {
-        SyncItem next = waitingTiles.pop_front();
-        if (_stop)
-           break;
-
-        SyncItem::Status cacheStatus = isPathCached(next);
-        if (cacheStatus != SyncItem::Invalid) {
-            _cache_hits++;
-            SG_LOG(SG_TERRASYNC, SG_DEBUG,
-                   "Cache hit for: '" << next._dir << "'");
-            next._status = cacheStatus;
-            _freshTiles.push_back(next);
-            _is_dirty = true;
-            continue;
-        }
-
-        syncPathExternal(next);
-
-        if ((_allowed_errors >= 0)&&
-            (_consecutive_errors >= _allowed_errors))
-        {
-            _stalled = true;
-            _stop = true;
-        }
-    } // of thread running loop
-}
-
-void SGTerraSync::SvnThread::syncPathExternal(const SyncItem& next)
-{
-    _busy = true;
-    SGPath path( _local_dir );
-    path.append( next._dir );
-    bool isNewDirectory = !path.exists();
-
-    try {
-        if (isNewDirectory) {
-            int rc = path.create_dir( 0755 );
-            if (rc) {
-                SG_LOG(SG_TERRASYNC,SG_ALERT,
-                       "Cannot create directory '" << path << "', return code = " << rc );
-                throw sg_exception("Cannot create directory for terrasync", path.str());
-            }
-        }
-
-        if (!runExternalSyncCommand(next._dir.c_str())) {
-            throw sg_exception("Running external sync command failed");
-        }
-    } catch (sg_exception& e) {
-        fail(next);
-        _busy = false;
-        return;
+        SGGuard<SGMutex> g(_stateLock);
+        _running = false;
     }
 
-    updated(next, isNewDirectory);
-    _busy = false;
 }
 
-void SGTerraSync::SvnThread::updateSyncSlot(SyncSlot &slot)
+void SGTerraSync::WorkerThread::updateSyncSlot(SyncSlot &slot)
 {
     if (slot.repository.get()) {
         if (slot.repository->isDoingSync()) {
@@ -676,14 +516,16 @@ void SGTerraSync::SvnThread::updateSyncSlot(SyncSlot &slot)
                 slot.nextWarnTimeout += 10000;
             }
 #endif
+            // convert bytes to kbytes here
+            slot.pendingKBytes = (slot.repository->bytesToDownload() >> 10);
             return; // easy, still working
         }
 
         // check result
-        SVNRepository::ResultCode res = slot.repository->failure();
-        if (res == AbstractRepository::REPO_ERROR_NOT_FOUND) {
+        HTTPRepository::ResultCode res = slot.repository->failure();
+        if (res == HTTPRepository::REPO_ERROR_NOT_FOUND) {
             notFound(slot.currentItem);
-        } else if (res != AbstractRepository::REPO_NO_ERROR) {
+        } else if (res != HTTPRepository::REPO_NO_ERROR) {
             fail(slot.currentItem);
         } else {
             updated(slot.currentItem, slot.isNewDirectory);
@@ -694,6 +536,7 @@ void SGTerraSync::SvnThread::updateSyncSlot(SyncSlot &slot)
         // whatever happened, we're done with this repository instance
         slot.busy = false;
         slot.repository.reset();
+        slot.pendingKBytes = 0;
     }
 
     // init and start sync of the next repository
@@ -714,18 +557,16 @@ void SGTerraSync::SvnThread::updateSyncSlot(SyncSlot &slot)
             }
         } // of creating directory step
 
-        string serverUrl(_svn_server);
-        if (!_httpServer.empty()) {
-          slot.repository.reset(new HTTPRepository(path, &_http));
-          serverUrl = _httpServer;
-        } else {
-          if (slot.currentItem._type == SyncItem::AIData) {
-              serverUrl = _svn_data_server;
-          }
-          slot.repository.reset(new SVNRepository(path, &_http));
+        slot.repository.reset(new HTTPRepository(path, &_http));
+        slot.repository->setEntireRepositoryMode();
+        slot.repository->setBaseUrl(_httpServer + "/" + slot.currentItem._dir);
+
+        if (_installRoot.exists()) {
+            SGPath p = _installRoot;
+            p.append(slot.currentItem._dir);
+            slot.repository->setInstalledCopyPath(p);
         }
 
-        slot.repository->setBaseUrl(serverUrl + "/" + slot.currentItem._dir);
         try {
             slot.repository->update();
         } catch (sg_exception& e) {
@@ -740,22 +581,27 @@ void SGTerraSync::SvnThread::updateSyncSlot(SyncSlot &slot)
         slot.nextWarnTimeout = 20000;
         slot.stamp.stamp();
         slot.busy = true;
+        slot.pendingKBytes = slot.repository->bytesToDownload();
+
         SG_LOG(SG_TERRASYNC, SG_INFO, "sync of " << slot.repository->baseUrl() << " started, queue size is " << slot.queue.size());
     }
 }
 
-void SGTerraSync::SvnThread::runInternal()
+void SGTerraSync::WorkerThread::runInternal()
 {
     while (!_stop) {
         try {
-            _http.update(100);
+            _http.update(10);
         } catch (sg_exception& e) {
             SG_LOG(SG_TERRASYNC, SG_WARN, "failure doing HTTP update" << e.getFormattedMessage());
         }
 
-        _transfer_rate = _http.transferRateBytesPerSec();
-        // convert from bytes to kbytes
-        _total_kb_downloaded = static_cast<int>(_http.totalBytesDownloaded() / 1024);
+        {
+            SGGuard<SGMutex> g(_stateLock);
+            _state._transfer_rate = _http.transferRateBytesPerSec();
+            // convert from bytes to kbytes
+            _state._total_kb_downloaded = static_cast<int>(_http.totalBytesDownloaded() / 1024);
+        }
 
         if (_stop)
             break;
@@ -765,11 +611,10 @@ void SGTerraSync::SvnThread::runInternal()
             SyncItem next = waitingTiles.pop_front();
             SyncItem::Status cacheStatus = isPathCached(next);
             if (cacheStatus != SyncItem::Invalid) {
-                _cache_hits++;
+                incrementCacheHits();
                 SG_LOG(SG_TERRASYNC, SG_DEBUG, "\nTerraSync Cache hit for: '" << next._dir << "'");
                 next._status = cacheStatus;
                 _freshTiles.push_back(next);
-                _is_dirty = true;
                 continue;
             }
 
@@ -778,13 +623,21 @@ void SGTerraSync::SvnThread::runInternal()
         }
 
         bool anySlotBusy = false;
+        unsigned int newPendingCount = 0;
+
         // update each sync slot in turn
         for (unsigned int slot=0; slot < NUM_SYNC_SLOTS; ++slot) {
             updateSyncSlot(_syncSlots[slot]);
+            newPendingCount += _syncSlots[slot].pendingKBytes;
             anySlotBusy |= _syncSlots[slot].busy;
         }
 
-        _busy = anySlotBusy;
+        {
+            SGGuard<SGMutex> g(_stateLock);
+            _state._totalKbPending = newPendingCount; // approximately atomic update
+            _state._busy = anySlotBusy;
+        }
+
         if (!anySlotBusy) {
             // wait on the blocking deque here, otherwise we spin
             // the loop very fast, since _http::update with no connections
@@ -794,7 +647,7 @@ void SGTerraSync::SvnThread::runInternal()
     } // of thread running loop
 }
 
-SyncItem::Status SGTerraSync::SvnThread::isPathCached(const SyncItem& next) const
+SyncItem::Status SGTerraSync::WorkerThread::isPathCached(const SyncItem& next) const
 {
     TileAgeCache::const_iterator ii = _completedTiles.find( next._dir );
     if (ii == _completedTiles.end()) {
@@ -816,20 +669,20 @@ SyncItem::Status SGTerraSync::SvnThread::isPathCached(const SyncItem& next) cons
     return (ii->second > now) ? SyncItem::Cached : SyncItem::Invalid;
 }
 
-void SGTerraSync::SvnThread::fail(SyncItem failedItem)
+void SGTerraSync::WorkerThread::fail(SyncItem failedItem)
 {
+    SGGuard<SGMutex> g(_stateLock);
     time_t now = time(0);
-    _consecutive_errors++;
-    _fail_count++;
+    _state._consecutive_errors++;
+    _state._fail_count++;
     failedItem._status = SyncItem::Failed;
     _freshTiles.push_back(failedItem);
     SG_LOG(SG_TERRASYNC,SG_INFO,
            "Failed to sync'" << failedItem._dir << "'");
     _completedTiles[ failedItem._dir ] = now + UpdateInterval::FailedAttempt;
-    _is_dirty = true;
 }
 
-void SGTerraSync::SvnThread::notFound(SyncItem item)
+void SGTerraSync::WorkerThread::notFound(SyncItem item)
 {
     // treat not found as authorative, so use the same cache expiry
     // as succesful download. Important for MP models and similar so
@@ -839,31 +692,33 @@ void SGTerraSync::SvnThread::notFound(SyncItem item)
     time_t now = time(0);
     item._status = SyncItem::NotFound;
     _freshTiles.push_back(item);
-    _is_dirty = true;
     _notFoundItems[ item._dir ] = now + UpdateInterval::SuccessfulAttempt;
     writeCompletedTilesPersistentCache();
 }
 
-void SGTerraSync::SvnThread::updated(SyncItem item, bool isNewDirectory)
+void SGTerraSync::WorkerThread::updated(SyncItem item, bool isNewDirectory)
 {
-    time_t now = time(0);
-    _consecutive_errors = 0;
-    _success_count++;
-    SG_LOG(SG_TERRASYNC,SG_INFO,
-           "Successfully synchronized directory '" << item._dir << "'");
+    {
+        SGGuard<SGMutex> g(_stateLock);
+        time_t now = time(0);
+        _state._consecutive_errors = 0;
+        _state._success_count++;
+        SG_LOG(SG_TERRASYNC,SG_INFO,
+               "Successfully synchronized directory '" << item._dir << "'");
 
-    item._status = SyncItem::Updated;
-    if (item._type == SyncItem::Tile) {
-        _updated_tile_count++;
+        item._status = SyncItem::Updated;
+        if (item._type == SyncItem::Tile) {
+            _state._updated_tile_count++;
+        }
+
+        _freshTiles.push_back(item);
+        _completedTiles[ item._dir ] = now + UpdateInterval::SuccessfulAttempt;
     }
 
-    _freshTiles.push_back(item);
-    _completedTiles[ item._dir ] = now + UpdateInterval::SuccessfulAttempt;
-    _is_dirty = true;
     writeCompletedTilesPersistentCache();
 }
 
-void SGTerraSync::SvnThread::initCompletedTilesPersistentCache()
+void SGTerraSync::WorkerThread::initCompletedTilesPersistentCache()
 {
     if (!_persistentCachePath.exists()) {
         return;
@@ -873,7 +728,7 @@ void SGTerraSync::SvnThread::initCompletedTilesPersistentCache()
     time_t now = time(0);
 
     try {
-        readProperties(_persistentCachePath.str(), cacheRoot);
+        readProperties(_persistentCachePath, cacheRoot);
     } catch (sg_exception& e) {
         SG_LOG(SG_TERRASYNC, SG_INFO, "corrupted persistent cache, discarding");
         return;
@@ -896,14 +751,14 @@ void SGTerraSync::SvnThread::initCompletedTilesPersistentCache()
     }
 }
 
-void SGTerraSync::SvnThread::writeCompletedTilesPersistentCache() const
+void SGTerraSync::WorkerThread::writeCompletedTilesPersistentCache() const
 {
     // cache is disabled
     if (_persistentCachePath.isNull()) {
         return;
     }
 
-    std::ofstream f(_persistentCachePath.c_str(), std::ios::trunc);
+    sg_ofstream f(_persistentCachePath, std::ios::trunc);
     if (!f.is_open()) {
         return;
     }
@@ -931,11 +786,11 @@ void SGTerraSync::SvnThread::writeCompletedTilesPersistentCache() const
 // SGTerraSync ////////////////////////////////////////////////////////////////
 ///////////////////////////////////////////////////////////////////////////////
 SGTerraSync::SGTerraSync() :
-    _svnThread(NULL),
+    _workerThread(NULL),
     _bound(false),
     _inited(false)
 {
-    _svnThread = new SvnThread();
+    _workerThread = new WorkerThread();
     _log = new BufferedLogCallback(SG_TERRASYNC, SG_INFO);
     _log->truncateAt(255);
 
@@ -944,8 +799,8 @@ SGTerraSync::SGTerraSync() :
 
 SGTerraSync::~SGTerraSync()
 {
-    delete _svnThread;
-    _svnThread = NULL;
+    delete _workerThread;
+    _workerThread = NULL;
     sglog().removeCallback(_log);
     delete _log;
      _tiedProperties.Untie();
@@ -965,56 +820,50 @@ void SGTerraSync::init()
     _inited = true;
 
     assert(_terraRoot);
-    _terraRoot->setBoolValue("built-in-svn-available",svn_built_in_available);
+    _terraRoot->setBoolValue("built-in-svn-available", true);
 
     reinit();
 }
 
 void SGTerraSync::shutdown()
 {
-     _svnThread->stop();
+     _workerThread->stop();
 }
 
 void SGTerraSync::reinit()
 {
     // do not reinit when enabled and we're already up and running
-    if ((_terraRoot->getBoolValue("enabled",false))&&
-         (_svnThread->_active && _svnThread->_running))
+    if ((_terraRoot->getBoolValue("enabled",false)) && _workerThread->isRunning())
     {
         return;
     }
 
-    _svnThread->stop();
+    _workerThread->stop();
 
     if (_terraRoot->getBoolValue("enabled",false))
     {
-        _svnThread->setSvnServer(_terraRoot->getStringValue("svn-server",""));
-        std::string httpServer(_terraRoot->getStringValue("http-server",""));
-        _svnThread->setHTTPServer(httpServer);
-        _svnThread->setSvnDataServer(_terraRoot->getStringValue("svn-data-server",""));
-        _svnThread->setRsyncServer(_terraRoot->getStringValue("rsync-server",""));
-        _svnThread->setLocalDir(_terraRoot->getStringValue("scenery-dir",""));
-        _svnThread->setAllowedErrorCount(_terraRoot->getIntValue("max-errors",5));
-        _svnThread->setUseBuiltin(_terraRoot->getBoolValue("use-built-in-svn",true));
+        _workerThread->setHTTPServer( _terraRoot->getStringValue("http-server","") );
+        _workerThread->setSceneryVersion( _terraRoot->getStringValue("scenery-version","ws20") );
+#if 1
+        // leave it hardcoded for now, not sure about the security implications for now
+        _workerThread->setDNSDN( "terrasync.flightgear.org");
+#else
+        _workerThread->setDNSDN( _terraRoot->getStringValue("dnsdn","terrasync.flightgear.org") );
+#endif
+        _workerThread->setLocalDir(_terraRoot->getStringValue("scenery-dir",""));
 
-        if (httpServer.empty()) {
-            // HTTP doesn't benefit from using the persistent cache
-            _svnThread->setCachePath(SGPath(_terraRoot->getStringValue("cache-path","")));
-        } else {
-            SG_LOG(SG_TERRASYNC, SG_INFO, "HTTP repository selected, disabling persistent cache");
-        }
+        SGPath installPath(_terraRoot->getStringValue("installation-dir"));
+        _workerThread->setInstalledDir(installPath);
+        _workerThread->setAllowedErrorCount(_terraRoot->getIntValue("max-errors",5));
+        _workerThread->setCacheHits(_terraRoot->getIntValue("cache-hit", 0));
 
-        _svnThread->setCacheHits(_terraRoot->getIntValue("cache-hit", 0));
-        _svnThread->setUseSvn(_terraRoot->getBoolValue("use-svn",true));
-        _svnThread->setExtSvnUtility(_terraRoot->getStringValue("ext-svn-utility","svn"));
-
-        if (_svnThread->start())
+        if (_workerThread->start())
         {
             syncAirportsModels();
         }
     }
 
-    _stalledNode->setBoolValue(_svnThread->_stalled);
+    _stalledNode->setBoolValue(_workerThread->isStalled());
 }
 
 void SGTerraSync::bind()
@@ -1024,17 +873,6 @@ void SGTerraSync::bind()
     }
 
     _bound = true;
-    _tiedProperties.Tie( _terraRoot->getNode("busy", true), (bool*) &_svnThread->_busy );
-    _tiedProperties.Tie( _terraRoot->getNode("active", true), (bool*) &_svnThread->_active );
-    _tiedProperties.Tie( _terraRoot->getNode("update-count", true), (int*) &_svnThread->_success_count );
-    _tiedProperties.Tie( _terraRoot->getNode("error-count", true), (int*) &_svnThread->_fail_count );
-    _tiedProperties.Tie( _terraRoot->getNode("tile-count", true), (int*) &_svnThread->_updated_tile_count );
-    _tiedProperties.Tie( _terraRoot->getNode("cache-hits", true), (int*) &_svnThread->_cache_hits );
-    _tiedProperties.Tie( _terraRoot->getNode("transfer-rate-bytes-sec", true), (int*) &_svnThread->_transfer_rate );
-
-    // use kbytes here because propety doesn't support 64-bit and we might conceivably
-    // download more than 2G in a single session
-    _tiedProperties.Tie( _terraRoot->getNode("downloaded-kbytes", true), (int*) &_svnThread->_total_kb_downloaded );
 
     _terraRoot->getNode("busy", true)->setAttribute(SGPropertyNode::WRITE,false);
     _terraRoot->getNode("active", true)->setAttribute(SGPropertyNode::WRITE,false);
@@ -1043,15 +881,16 @@ void SGTerraSync::bind()
     _terraRoot->getNode("tile-count", true)->setAttribute(SGPropertyNode::WRITE,false);
     _terraRoot->getNode("use-built-in-svn", true)->setAttribute(SGPropertyNode::USERARCHIVE,false);
     _terraRoot->getNode("use-svn", true)->setAttribute(SGPropertyNode::USERARCHIVE,false);
+
     // stalled is used as a signal handler (to connect listeners triggering GUI pop-ups)
     _stalledNode = _terraRoot->getNode("stalled", true);
-    _stalledNode->setBoolValue(_svnThread->_stalled);
+    _stalledNode->setBoolValue(_workerThread->isStalled());
     _stalledNode->setAttribute(SGPropertyNode::PRESERVE,true);
 }
 
 void SGTerraSync::unbind()
 {
-    _svnThread->stop();
+    _workerThread->stop();
     _tiedProperties.Untie();
     _bound = false;
     _inited = false;
@@ -1063,37 +902,31 @@ void SGTerraSync::unbind()
 
 void SGTerraSync::update(double)
 {
-    static SGBucket bucket;
-    if (_svnThread->isDirty())
+    TerrasyncThreadState copiedState(_workerThread->threadsafeCopyState());
+
+    _terraRoot->setBoolValue("busy", copiedState._busy);
+    _terraRoot->setIntValue("update-count", copiedState._success_count);
+    _terraRoot->setIntValue("error-count", copiedState._fail_count);
+    _terraRoot->setIntValue("tile-count", copiedState._updated_tile_count);
+    _terraRoot->setIntValue("cache-hits", copiedState._cache_hits);
+    _terraRoot->setIntValue("transfer-rate-bytes-sec", copiedState._transfer_rate);
+    _terraRoot->setIntValue("downloaded-kbytes", copiedState._total_kb_downloaded);
+    _terraRoot->setIntValue("pending-kbytes", copiedState._totalKbPending);
+
+    _stalledNode->setBoolValue(_workerThread->isStalled());
+    _terraRoot->setBoolValue("active", _workerThread->isRunning());
+
+    while (_workerThread->hasNewTiles())
     {
-        if (!_svnThread->_active)
-        {
-            if (_svnThread->_stalled)
-            {
-                SG_LOG(SG_TERRASYNC,SG_ALERT,
-                       "Automatic scenery download/synchronization stalled. Too many errors.");
-            }
-            else
-            {
-                // not really an alert - just always show this message
-                SG_LOG(SG_TERRASYNC,SG_ALERT,
-                        "Automatic scenery download/synchronization has stopped.");
-            }
-            _stalledNode->setBoolValue(_svnThread->_stalled);
+        SyncItem next = _workerThread->getNewTile();
+
+        if ((next._type == SyncItem::Tile) || (next._type == SyncItem::AIData)) {
+            _activeTileDirs.erase(next._dir);
         }
-
-        while (_svnThread->hasNewTiles())
-        {
-            SyncItem next = _svnThread->getNewTile();
-
-            if ((next._type == SyncItem::Tile) || (next._type == SyncItem::AIData)) {
-                _activeTileDirs.erase(next._dir);
-            }
-        } // of freshly synced items
-    }
+    } // of freshly synced items
 }
 
-bool SGTerraSync::isIdle() {return _svnThread->isIdle();}
+bool SGTerraSync::isIdle() {return _workerThread->isIdle();}
 
 void SGTerraSync::syncAirportsModels()
 {
@@ -1106,12 +939,12 @@ void SGTerraSync::syncAirportsModels()
             ostringstream dir;
             dir << "Airports/" << synced_other;
             SyncItem w(dir.str(), SyncItem::AirportData);
-            _svnThread->request( w );
+            _workerThread->request( w );
         }
     }
 
     SyncItem w("Models", SyncItem::SharedModels);
-    _svnThread->request( w );
+    _workerThread->request( w );
 }
 
 void SGTerraSync::syncAreaByPath(const std::string& aPath)
@@ -1126,7 +959,7 @@ void SGTerraSync::syncAreaByPath(const std::string& aPath)
 
         _activeTileDirs.insert(dir);
         SyncItem w(dir, SyncItem::Tile);
-        _svnThread->request( w );
+        _workerThread->request( w );
     }
 }
 
@@ -1139,7 +972,7 @@ bool SGTerraSync::scheduleTile(const SGBucket& bucket)
 
 bool SGTerraSync::isTileDirPending(const std::string& sceneryDir) const
 {
-    if (!_svnThread->_running) {
+    if (!_workerThread->isRunning()) {
         return false;
     }
 
@@ -1162,13 +995,13 @@ void SGTerraSync::scheduleDataDir(const std::string& dataDir)
 
     _activeTileDirs.insert(dataDir);
     SyncItem w(dataDir, SyncItem::AIData);
-    _svnThread->request( w );
+    _workerThread->request( w );
 
 }
 
 bool SGTerraSync::isDataDirPending(const std::string& dataDir) const
 {
-    if (!_svnThread->_running) {
+    if (!_workerThread->isRunning()) {
         return false;
     }
 
